import logging
from contextlib import asynccontextmanager
from typing import Mapping
from typing import Optional

import aiozipkin as az
from aiozipkin import make_context
from aiozipkin.helpers import Endpoint
from aiozipkin.helpers import TraceContext

from easyjaeger.vendor import EasyTracer

from ..vendor.tracer import create
from .config import SINGLE_HEADER
from .config import TracingConfig


@asynccontextmanager
async def append_trace(
    config: TracingConfig,
    span_name: str,
    span_start: str,
    span_end: str,
    span_kind: str,
    span_tags: Optional[Mapping[str, str]] = None,
    parent_id: Optional[str] = None,
) -> str:
    endpoint: Endpoint = az.create_endpoint(
        config.service_name, ipv4=config.service_host, port=config.service_port
    )
    tracer: EasyTracer = await create(
        config.host, endpoint, sample_rate=config.sample_rate
    )
    tracer_context: Optional[TraceContext] = make_context({'b3': parent_id})
    if not tracer_context:
        # TODO: Handle if tracer_context is None
        logging.warning(f'Incorrect parent_id. Ignoring span')
        yield None
    try:
        if tracer_context:
            with tracer.new_child(tracer_context) as span:
                span.name(span_name)
                span.kind(span_kind)
                span.annotate(span_start)
                if span_tags:
                    for tag in span_tags.items():
                        span.tag(*tag)
                try:
                    yield span.context.make_single_header()[SINGLE_HEADER]
                finally:
                    span.annotate(span_end)
    finally:
        await tracer.close()


__all__ = ['append_trace']



from dataclasses import dataclass


@dataclass
class TracingConfig:
    host: str
    service_name: str
    service_host: str
    service_port: int
    sample_rate: float = 1.0
    sampled: bool = True


# possible span kinds
CLIENT = 'CLIENT'
SERVER = 'SERVER'
PRODUCER = 'PRODUCER'
CONSUMER = 'CONSUMER'

SINGLE_HEADER = 'b3'

__all__ = ['TracingConfig', 'CLIENT', 'SERVER', 'PRODUCER', 'CONSUMER', 'SINGLE_HEADER']



from contextlib import asynccontextmanager
from typing import Mapping
from typing import Optional

import aiozipkin as az

from ..vendor.tracer import create
from .config import SINGLE_HEADER
from .config import TracingConfig


@asynccontextmanager
async def init_trace(
    config: TracingConfig,
    span_name: str,
    span_start: str,
    span_end: str,
    span_kind: str,
    span_tags: Optional[Mapping[str, str]] = None,
    trace_id: Optional[str] = None,
) -> str:
    endpoint = az.create_endpoint(
        config.service_name, ipv4=config.service_host, port=config.service_port
    )
    tracer = await create(config.host, endpoint, sample_rate=config.sample_rate)
    try:
        with tracer.new_trace(trace_id, config.sampled) as span:
            span.name(span_name)
            span.kind(span_kind)
            span.annotate(span_start)
            if span_tags:
                for tag in span_tags.items():
                    span.tag(*tag)
            try:
                yield span.context.make_single_header()[SINGLE_HEADER]
            finally:
                span.annotate(span_end)
    finally:
        await tracer.close()


__all__ = ['init_trace']



import logging
from typing import Optional

import aiozipkin as az
from aiohttp.web import Request
from aiohttp.web import Response
from aiohttp.web_middlewares import middleware
from aiozipkin.helpers import TraceContext
from aiozipkin.helpers import make_context

from ...tracer.config import SINGLE_HEADER
from ...vendor.tracer import create
from .middleware_config import MiddlewareConfig


def child_middleware(config: MiddlewareConfig,):
    @middleware
    async def trace_middleware(request: Request, handler):
        parent_id: str = request.headers.get('Easy-Trace-ID', 'unknown')
        tracer_context: Optional[TraceContext] = make_context({'b3': parent_id})
        if not tracer_context:
            logging.warning(f'Incorrect parent_id. Ignoring middleware')
            return await handler(request)

        endpoint = az.create_endpoint(
            config.tracing_config.service_name,
            ipv4=config.tracing_config.service_host,
            port=config.tracing_config.service_port,
        )
        tracer = await create(
            config.tracing_config.host,
            endpoint,
            sample_rate=config.tracing_config.sample_rate,
        )
        try:
            with tracer.new_child(tracer_context) as span:
                span.name(config.span_name)
                span.kind(config.span_kind)
                span.annotate(config.span_start)
                if config.span_tags:
                    for tag in config.span_tags.items():
                        span.tag(*tag)
                try:
                    request[
                        config.request_trace_id_name
                    ] = span.context.make_single_header()[SINGLE_HEADER]
                    response: Response = await handler(request)
                    response.headers['Easy-Trace-ID'] = request[
                        config.request_trace_id_name
                    ]
                    return response
                finally:
                    span.annotate(config.span_end)
        finally:
            await tracer.close()

    return trace_middleware


__all__ = ['child_middleware']



from dataclasses import dataclass
from typing import Mapping
from typing import Optional

from ...tracer.config import SERVER
from ...tracer.config import TracingConfig


@dataclass
class MiddlewareConfig:
    tracing_config: TracingConfig
    trace_id: Optional[str] = None
    request_trace_id_name: str = 'easy_trace_id'
    skip_routes: tuple = ()
    span_name: str = 'Entrypoint'
    span_start: str = 'Start request'
    span_end: str = 'End request'
    span_kind: str = SERVER
    span_tags: Optional[Mapping[str, str]] = None


__all__ = ['MiddlewareConfig']



from typing import Awaitable
from typing import Optional

from aiozipkin import Sampler
from aiozipkin import Tracer
from aiozipkin.context_managers import _ContextManager
from aiozipkin.helpers import Endpoint
from aiozipkin.helpers import TraceContext
from aiozipkin.mypy_types import OptBool
from aiozipkin.mypy_types import OptLoop
from aiozipkin.span import SpanAbc
from aiozipkin.transport import Transport
from aiozipkin.utils import generate_random_64bit_string
from aiozipkin.utils import generate_random_128bit_string


class EasyTracer(Tracer):
    def _next_context(
        self,
        context: Optional[TraceContext] = None,
        sampled: OptBool = None,
        debug: bool = False,
        trace_id: Optional[str] = None,
    ) -> TraceContext:
        span_id = generate_random_64bit_string()
        if context is not None:
            new_context = context._replace(
                span_id=span_id, parent_id=context.span_id, shared=False
            )
            return new_context

        trace_id = trace_id if trace_id else generate_random_128bit_string()
        if sampled is None:
            sampled = self._sampler.is_sampled(trace_id)

        new_context = TraceContext(
            trace_id=trace_id,
            parent_id=None,
            span_id=span_id,
            sampled=sampled,
            debug=debug,
            shared=False,
        )
        return new_context

    def new_trace(
        self,
        trace_id: Optional[str] = None,
        sampled: OptBool = None,
        debug: bool = False,
    ) -> SpanAbc:
        context = self._next_context(
            None, sampled=sampled, debug=debug, trace_id=trace_id
        )
        return self.to_span(context)


def create(
    zipkin_address: str,
    local_endpoint: Endpoint,
    *,
    sample_rate: float = 0.01,
    send_interval: float = 5,
    loop: OptLoop = None
) -> Awaitable[EasyTracer]:
    async def build_tracer() -> EasyTracer:
        sampler = Sampler(sample_rate=sample_rate)
        transport = Transport(zipkin_address, send_interval=send_interval, loop=loop)
        return EasyTracer(transport, sampler, local_endpoint)

    result = _ContextManager(build_tracer())
    return result


__all__ = ['create', 'EasyTracer']



from .child_trace import append_trace
from .config import CLIENT
from .config import CONSUMER
from .config import PRODUCER
from .config import SERVER
from .config import TracingConfig
from .parent_trace import init_trace

__all__ = [
    'TracingConfig',
    'CLIENT',
    'CONSUMER',
    'SERVER',
    'PRODUCER',
    'init_trace',
    'append_trace',
]



from .helpers import aiohttp_helpers
from .tracer import CLIENT
from .tracer import CONSUMER
from .tracer import PRODUCER
from .tracer import SERVER
from .tracer import TracingConfig
from .tracer import append_trace
from .tracer import init_trace

__all__ = [
    'TracingConfig',
    'CLIENT',
    'CONSUMER',
    'SERVER',
    'PRODUCER',
    'init_trace',
    'append_trace',
    'aiohttp_helpers',
]



from .child_middleware import child_middleware
from .middleware_config import MiddlewareConfig
from .parent_middleware import parent_middleware

__all__ = ['MiddlewareConfig', 'parent_middleware', 'child_middleware']



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    'aiozipkin==0.6.0', 'aiohttp'
]

test_require = {
    'pytest',
    'pytest-asyncio',
    'aiozipkin==0.6.0'
}

setup(
    name='easyjaeger',
    version='0.3.0',
    description='Helper for easy breezy jaeger/zipkin in asyncio python services',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/borisrozumnuk/cognitojwt',
    author='Michail Shatalov',
    author_email='misha.shatalow@gmail.com',
    license='MIT',
    platforms='Any',
    install_requires=install_requires,
    keywords='zipkin jaeger tracing aiozipkin',
    packages=find_packages(exclude=('tests')),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
    ],
    zip_safe=False
)
